"""Provide the ImJoy core API interface."""
from enum import Enum
from contextvars import ContextVar
from typing import Any, Callable, Dict, List, Optional

from pydantic import (  # pylint: disable=no-name-in-module
    BaseModel,
    EmailStr,
    PrivateAttr,
)


class TokenConfig(BaseModel):
    """Represent a token configuration."""

    scopes: List[str]
    expires_in: Optional[int]
    email: Optional[EmailStr]


class VisibilityEnum(str, Enum):
    """Represent the visibility of the workspace."""

    public = "public"
    protected = "protected"


class UserInfo(BaseModel):
    """Represent user info."""

    id: str
    roles: List[str]
    email: Optional[EmailStr]
    parent: Optional[str]
    scopes: Optional[List[str]]  # a list of workspace
    expires_at: Optional[int]
    _plugins: Dict[str, Any] = PrivateAttr(default_factory=lambda: {})  # id:plugin
    _sessions: List[str] = PrivateAttr(default_factory=lambda: [])  # session ids


class WorkspaceInfo(BaseModel):
    """Represent a workspace."""

    name: str
    persistent: bool
    owners: List[str]
    visibility: VisibilityEnum
    description: Optional[str]
    icon: Optional[str]
    covers: Optional[List[str]]
    docs: Optional[str]
    allow_list: Optional[List[str]]
    deny_list: Optional[List[str]]
    authorizer: Optional[str]
    _authorizer: Optional[Callable] = PrivateAttr(default_factory=lambda: None)
    _plugins: Dict[str, Any] = PrivateAttr(default_factory=lambda: {})  # name: plugin
    _services: List[Dict[str, Any]] = PrivateAttr(default_factory=lambda: [])


current_user = ContextVar("current_user")
current_plugin = ContextVar("current_plugin")
current_workspace = ContextVar("current_workspace")
all_sessions: Dict[str, UserInfo] = {}  # sid:user_info
all_users: Dict[str, UserInfo] = {}  # uid:user_info
all_workspaces: Dict[str, WorkspaceInfo] = {}  # wid:workspace_info



"""Provide the connection."""
import asyncio
import logging
import sys
import time

from imjoy_rpc.utils import MessageEmitter, dotdict

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("core-connection")
logger.setLevel(logging.WARNING)

all_connections = {}


class BasicConnection(MessageEmitter):
    """Represent a base connection."""

    def __init__(self, send):
        """Set up instance."""
        super().__init__(logger)
        self.plugin_config = dotdict()
        self._send = send
        self._access_token = None
        self._expires_in = None
        self._plugin_origin = "*"
        self._refresh_token = None
        self.peer_id = None
        self.on("initialized", self._initialized)

    def _initialized(self, data):
        self.plugin_config = data["config"]
        # peer_id can only be set for once
        self.peer_id = data["peer_id"]
        self._plugin_origin = data.get("origin", "*")
        all_connections[self.peer_id] = self
        if self._plugin_origin != "*":
            logger.info(
                "Connection to the imjoy-rpc peer $%s is limited to origin %s.",
                self.peer_id,
                self._plugin_origin,
            )

        if not self.peer_id:
            raise Exception("Please provide a peer_id for the connection.")

        if self.plugin_config.get("auth"):
            if self._plugin_origin == "*":
                logger.error(
                    "Refuse to transmit the token without an explicit origin, "
                    "there is a security risk that you may leak the credential "
                    "to website from other origin. "
                    "Please specify the `origin` explicitly."
                )
                self._access_token = None
                self._refresh_token = None

            if self.plugin_config["auth"]["type"] != "jwt":
                logger.error(
                    "Unsupported authentication type: %s", self.plugin_config.auth.type
                )
            else:
                self._expires_in = self.plugin_config["auth"]["expires_in"]
                self._access_token = self.plugin_config["auth"]["access_token"]
                self._refresh_token = self.plugin_config["auth"]["refresh_token"]

    def handle_message(self, data):
        """Handle a message."""
        target_id = data.get("target_id")
        if target_id and self.peer_id and target_id != self.peer_id:
            conn = all_connections[target_id]
            if conn:
                conn._fire(data.type, data)  # pylint: disable=protected-access
            else:
                logger.warning(
                    "Connection with target_id %s not found, discarding data: %s",
                    target_id,
                    data,
                )
        else:
            self._fire(data["type"], data)

    def connect(self):
        """Connect."""
        self._fire("connected")

    async def execute(self, code):
        """Execute."""
        raise PermissionError

    def emit(self, msg):
        """Send a message to the plugin site."""
        if self._access_token:
            if time.time() >= self._expires_in:
                # TODO: refresh access token
                raise Exception("Refresh token is not implemented.")

            msg["access_token"] = self._access_token
        msg["peer_id"] = msg.get("peer_id") or self.peer_id
        asyncio.ensure_future(self._send(msg))

    def disconnect(self, details):
        """Disconnect the plugin."""
        if self.peer_id and self.peer_id in all_connections:
            del all_connections[self.peer_id]



"""Provide an example plugin."""
from imjoy_rpc import api

# pylint: disable=no-self-use


class ImJoyPlugin:
    """Represent an ImJoy plugin."""

    def authorizer(self, user_info, context):
        """Authorize users."""
        return True

    async def setup(self):
        """Set up the plugin."""
        token = await api.generateToken()
        assert token["token"]
        print(f'Generated token: {token["token"]}')

        await api.register_service(
            {"name": "echo service", "echo": lambda x: print("echo: " + str(x))}
        )
        service = await api.get_services({"name": "echo service"})
        await service[0].echo("a message")
        await api.log("initialized")

    async def run(self, ctx):
        """Run the plugin."""
        await api.log("hello world")


api.export(ImJoyPlugin(), config={"name": "test-plugin", "workspace": "123"})



"""Provide the server."""
import asyncio
import os
import uuid
from contextvars import copy_context
from os import environ as env
from typing import Union

import socketio
import uvicorn
from dotenv import find_dotenv, load_dotenv
from fastapi import FastAPI
from fastapi.logger import logger
from fastapi.middleware.cors import CORSMiddleware

from imjoy import __version__ as VERSION
from imjoy.core import (
    UserInfo,
    VisibilityEnum,
    WorkspaceInfo,
    all_users,
    all_sessions,
    current_user,
    current_plugin,
    current_workspace,
    all_workspaces,
)
from imjoy.core.auth import parse_token, check_permission
from imjoy.core.connection import BasicConnection
from imjoy.core.interface import CoreInterface
from imjoy.core.plugin import DynamicPlugin

ENV_FILE = find_dotenv()
if ENV_FILE:
    load_dotenv(ENV_FILE)


def initialize_socketio(sio, core_api):
    """Initialize socketio."""
    # pylint: disable=too-many-statements, unused-variable, protected-access

    @sio.event
    async def connect(sid, environ):
        """Handle event called when a socketio client is connected to the server."""
        if "HTTP_AUTHORIZATION" in environ:
            try:
                authorization = environ["HTTP_AUTHORIZATION"]  # JWT token
                user_info = parse_token(authorization)
                uid = user_info["user_id"]
                email = user_info["email"]
                roles = user_info["roles"]
                parent = user_info.get("parent")
                scopes = user_info.get("scopes") or []
                expires_at = user_info.get("expires_at")
            except Exception as err:  # pylint: disable=broad-except
                logger.exception("Authentication failed: %s", err)
                # The connect event handler can return False
                # to reject the connection with the client.
                return False
            logger.info("User connected: %s", uid)
        else:
            uid = str(uuid.uuid4())
            email = None
            roles = []
            parent = None
            scopes = []
            expires_at = None
            logger.info("Anonymized User connected: %s", uid)

        if uid not in all_users:
            all_users[uid] = UserInfo(
                id=uid,
                email=email,
                parent=parent,
                roles=roles,
                scopes=scopes,
                expires_at=expires_at,
            )
        all_users[uid]._sessions.append(sid)
        all_sessions[sid] = all_users[uid]

    @sio.event
    async def register_plugin(sid, config):
        user_info = all_sessions[sid]
        ws = config.get("workspace") or user_info.id
        config["workspace"] = ws
        config["name"] = config.get("name") or str(uuid.uuid4())
        if ws in all_workspaces:
            workspace = all_workspaces[ws]
        else:
            if ws == user_info.id:
                # create the user workspace automatically
                workspace = WorkspaceInfo(
                    name=ws,
                    owners=[user_info.id],
                    visibility=VisibilityEnum.protected,
                    persistent=(config.get("persistent") is True),
                )
                all_workspaces[ws] = workspace
            else:
                return {"success": False, "detail": f"Workspace {ws} does not exist."}

        if user_info.id != ws and not check_permission(workspace, user_info):
            return {
                "success": False,
                "detail": f"Permission denied for workspace: {ws}",
            }

        name = config["name"].replace("/", "-")  # prevent hacking of the plugin name
        plugin_id = f"{ws}/{name}"
        config["id"] = plugin_id
        sio.enter_room(sid, plugin_id)

        async def send(data):
            await sio.emit(
                "plugin_message",
                data,
                room=plugin_id,
            )

        connection = BasicConnection(send)
        plugin = DynamicPlugin(config, core_api.get_interface(), connection, workspace)

        user_info._plugins[plugin.id] = plugin
        if plugin.name in workspace._plugins:
            # kill the plugin if already exist
            asyncio.ensure_future(plugin.terminate(True))
            del user_info._plugins[plugin.id]
        workspace._plugins[plugin.name] = plugin
        logger.info("New plugin registered successfully (%s)", plugin_id)
        return {"success": True, "plugin_id": plugin_id}

    @sio.event
    async def plugin_message(sid, data):
        user_info = all_sessions[sid]
        plugin_id = data["plugin_id"]
        ws, name = os.path.split(plugin_id)
        if ws not in all_workspaces:
            return {"success": False, "detail": f"Workspace not found: {ws}"}
        workspace = all_workspaces[ws]
        if user_info.id != ws and not check_permission(workspace, user_info):
            logger.error(
                "Permission denied: workspace=%s, user_id=%s", workspace, user_info.id
            )
            return {"success": False, "detail": "Permission denied"}

        plugin = workspace._plugins.get(name)
        if not plugin:
            logger.warning("Plugin %s not found in workspace %s", name, workspace.name)
            return {
                "success": False,
                "detail": f"Plugin {name} not found in workspace {workspace.name}",
            }

        current_user.set(user_info)
        current_plugin.set(plugin)
        current_workspace.set(workspace)
        ctx = copy_context()
        ctx.run(plugin.connection.handle_message, data)
        return {"success": True}

    @sio.event
    async def disconnect(sid):
        """Event handler called when the client is disconnected."""
        user_info = all_sessions[sid]
        all_users[user_info.id]._sessions.remove(sid)
        # if the user has no more all_sessions
        if not all_users[user_info.id]._sessions:
            del all_users[user_info.id]
            for pid in list(user_info._plugins.keys()):
                plugin = user_info._plugins[pid]
                # TODO: how to allow plugin running when the user disconnected
                # we will also need to handle the case when the user login again
                # the plugin should be reclaimed for the user
                del plugin.workspace._plugins[plugin.name]
                # if there is no plugins in the workspace then we remove it
                if not plugin.workspace._plugins and not plugin.workspace.persistent:
                    del all_workspaces[plugin.workspace.name]
                asyncio.ensure_future(plugin.terminate())
                del user_info._plugins[pid]

                # TODO: if a workspace has no plugins anymore
                # we should destroy it completely
                # Importantly, if we want to recycle the workspace name,
                # we need to make sure we don't mess up with the permission
                # with the plugins of the previous owners
                for service in plugin.workspace._services.copy():
                    if service.providerId == plugin.id:
                        plugin.workspace._services.remove(service)
        del all_sessions[sid]


def create_application(allow_origins) -> FastAPI:
    """Set up the server application."""
    # pylint: disable=unused-variable, protected-access

    app = FastAPI(
        title="ImJoy Core Server",
        description=(
            "A server for managing imjoy plugins and \
                enabling remote procedure calls"
        ),
        version=VERSION,
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=allow_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["Content-Type", "Authorization"],
    )

    @app.get("/")
    async def root():
        return {
            "name": "ImJoy Core Server",
            "version": VERSION,
            "all_users": {u: all_users[u]._sessions for u in all_users},
            "all_workspaces": {
                w.name: len(w._plugins) for w in all_workspaces.values()
            },
        }

    return app


def setup_socketio_server(
    app: FastAPI,
    mount_location: str = "/",
    socketio_path: str = "socket.io",
    allow_origins: Union[str, list] = "*",
) -> None:
    """Set up the socketio server."""
    if allow_origins == ["*"]:
        allow_origins = "*"
    sio = socketio.AsyncServer(async_mode="asgi", cors_allowed_origins=allow_origins)
    _app = socketio.ASGIApp(socketio_server=sio, socketio_path=socketio_path)

    app.mount(mount_location, _app)
    app.sio = sio
    core_api = CoreInterface()
    initialize_socketio(sio, core_api)
    return sio


def start_server(args):
    """Start the socketio server."""
    if args.allow_origin:
        allow_origin = args.allow_origin.split(",")
    else:
        allow_origin = env.get("ALLOW_ORIGINS", "*").split(",")
    application = create_application(allow_origin)
    setup_socketio_server(application, allow_origins=allow_origin)
    if args.host == "127.0.0.1" or args.host == "localhost":
        print(
            "***Note: If you want to enable access from another host,\
                 please start with `--host=0.0.0.0`.***"
        )
    uvicorn.run(application, host=args.host, port=int(args.port))


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--host",
        type=str,
        default="127.0.0.1",
        help="host for the socketio server",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=3000,
        help="port for the socketio server",
    )
    parser.add_argument(
        "--allow-origin",
        type=str,
        default="*",
        help="origins for the socketio server",
    )
    opt = parser.parse_args()
    start_server(opt)



"""Provide main entrypoint."""
import asyncio
import json
import logging
import os
import re
import sys
import urllib.request

import yaml
import imjoy_rpc
from imjoy_rpc import connect_to_server

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("plugin-runner")
logger.setLevel(logging.INFO)


async def run_plugin(plugin_file, default_config):
    """Load plugin file."""
    loop = asyncio.get_event_loop()
    if os.path.isfile(plugin_file):
        content = open(plugin_file).read()
    elif plugin_file.startswith("http"):
        with urllib.request.urlopen(plugin_file) as response:
            content = response.read().decode("utf-8")
        # remove query string
        plugin_file = plugin_file.split("?")[0]
    else:
        raise Exception("Invalid input plugin file path: {}".format(plugin_file))

    if plugin_file.endswith(".py"):
        filename, _ = os.path.splitext(os.path.basename(plugin_file))
        default_config["name"] = filename[:32]
        api = await connect_to_server(default_config)
        # patch imjoy_rpc api
        imjoy_rpc.api = api
        exec(content, globals())  # pylint: disable=exec-used
        logger.info("Plugin executed")
        if opt.quit_on_ready:
            await asyncio.sleep(1)
            loop.stop()

    elif plugin_file.endswith(".imjoy.html"):
        # load config
        found = re.findall("<config (.*)>\n(.*)</config>", content, re.DOTALL)[0]
        if "json" in found[0]:
            plugin_config = json.loads(found[1])
        elif "yaml" in found[0]:
            plugin_config = yaml.safe_load(found[1])
        default_config.update(plugin_config)
        api = await connect_to_server(default_config)
        # load script
        found = re.findall("<script (.*)>\n(.*)</script>", content, re.DOTALL)[0]
        if "python" in found[0]:
            # patch imjoy_rpc api
            imjoy_rpc.api = api
            exec(found[1], globals())  # pylint: disable=exec-used
            logger.info("Plugin executed")
            if opt.quit_on_ready:
                await asyncio.sleep(1)
                loop.stop()
        else:
            raise RuntimeError(
                "Invalid script type ({}) in file {}".format(found[0], plugin_file)
            )
    else:
        raise RuntimeError("Invalid script file type ({})".format(plugin_file))


async def start(args):
    """Run the plugin."""
    try:
        default_config = {
            "server_url": args.server_url,
            "workspace": args.workspace,
            "token": args.token,
        }
        await run_plugin(args.file, default_config)
    except Exception:  # pylint: disable=broad-except
        logger.exception("Failed to run plugin.")
        loop = asyncio.get_event_loop()
        loop.stop()
        sys.exit(1)


def start_runner(args):
    """Start the plugin runner."""
    loop = asyncio.get_event_loop()
    asyncio.ensure_future(start(args))
    loop.run_forever()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file", type=str, help="path to a plugin file")
    parser.add_argument(
        "--server-url",
        type=str,
        default=None,
        help="url to the plugin socketio server",
    )

    parser.add_argument(
        "--workspace",
        type=str,
        default=None,
        help="the plugin workspace",
    )

    parser.add_argument(
        "--token",
        type=str,
        default=None,
        help="token for the plugin workspace",
    )

    parser.add_argument(
        "--quit-on-ready",
        action="store_true",
        help="quit the server when the plugin is ready",
    )

    opt = parser.parse_args()

    start_runner(opt)



"""Provide setup function to prepare the engine."""
import argparse
import os

from imjoy import __version__


def parse_cmd_line(args=None):
    """Parse the command line options."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--jupyter", action="store_true", help="start jupyter notebook server"
    )
    parser.add_argument("--token", type=str, default=None, help="connection token")
    parser.add_argument(
        "--random-token", action="store_true", help="randomly generate a token"
    )
    parser.add_argument("--debug", action="store_true", help="debug mode")
    parser.add_argument(
        "--insecure",
        action="store_true",
        help="disable authentication token and xsrf. DO NOT USE IN PRODUCTION!!!",
    )
    parser.add_argument(
        "--serve",
        action="store_true",
        help="start the socketio server backend",
    )
    parser.add_argument("--host", type=str, default="127.0.0.1", help="server host")
    parser.add_argument(
        "--base-url",
        type=str,
        default=None,
        help="the base url for accessing this plugin engine",
    )
    parser.add_argument(
        "--allow-origin",
        type=str,
        default="*",
        help="the allow origin header to prevent unintended access from other website",
    )
    parser.add_argument(
        "--content-security-policy",
        type=str,
        default="frame-ancestors 'self' https://imjoy.io https://*.imjoy.io",
        help=(
            "the Content-Security-Policy header "
            "to prevent unintended access from other website"
        ),
    )
    parser.add_argument("--port", type=str, default="9527", help="server port")
    parser.add_argument(
        "--workspace",
        type=str,
        default=os.path.abspath(os.getcwd()),
        help="workspace folder for plugins",
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )

    opt = parser.parse_args(args=args)

    if opt.base_url is None or opt.base_url == "":
        opt.base_url = "http://{}:{}".format(opt.host, opt.port)

    if opt.base_url.endswith("/"):
        opt.base_url = opt.base_url[:-1]

    return opt



"""Provide main entrypoint."""
import logging
import sys

from imjoy.options import parse_cmd_line
from imjoy.utils import read_or_generate_token, write_token

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("main")
logger.setLevel(logging.INFO)


def main():
    """Run main."""
    opt = parse_cmd_line()
    if opt.serve:
        # pylint: disable=import-outside-toplevel
        from imjoy.server import start_server

        start_server(opt)
    elif opt.jupyter:
        sys.argv = sys.argv[:1]
        sys.argc = 1
        # pylint: disable=import-outside-toplevel
        from notebook.notebookapp import NotebookApp

        kwargs = {
            "open_browser": False,
            "allow_origin": opt.allow_origin,
            "ip": opt.host,
            "notebook_dir": opt.workspace,
            "port": int(opt.port),
            "tornado_settings": {
                "headers": {
                    "Access-Control-Allow-Origin": opt.allow_origin,
                    "Content-Security-Policy": opt.content_security_policy,
                }
            },
        }

        if not opt.token:
            if not opt.random_token:
                opt.token = read_or_generate_token()
                kwargs["token"] = opt.token
        else:
            kwargs["token"] = opt.token

        if opt.insecure:
            kwargs["token"] = ""
            kwargs["disable_check_xsrf"] = True
            logger.warning(
                "Running Jupyter notebooks with --insecure flag, "
                "please do not use it for production."
            )

        app = NotebookApp.instance(**kwargs)
        app.initialize()
        if app.port != int(opt.port):
            print("\nWARNING: using a different port: {}.\n".format(app.port))
        write_token(app.token)
        app._token_generated = True  # pylint: disable=protected-access
        app.start()

    else:
        print(
            "\nNote: We have migrated the backend of the ImJoy Engine."
            "You can start the Jupyter backend via`imjoy --jupyter`.\n\n"
            "Or, you can start the socketio backend server via `imjoy --serve`\n"
        )


if __name__ == "__main__":
    main()



"""Test the imjoy engine server."""
import os
import subprocess
import sys
import time

import pytest
import requests
from requests import RequestException
from imjoy_rpc import connect_to_server

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

PORT = 38283
SERVER_URL = f"http://127.0.0.1:{PORT}"


@pytest.fixture(name="socketio_server")
def socketio_server_fixture():
    """Start server as test fixture and tear down after test."""
    with subprocess.Popen(
        [sys.executable, "-m", "imjoy.server", f"--port={PORT}"]
    ) as proc:

        timeout = 5
        while timeout > 0:
            try:
                response = requests.get(f"http://127.0.0.1:{PORT}/")
                if response.ok:
                    break
            except RequestException:
                pass
            timeout -= 0.1
            time.sleep(0.1)
        yield

        proc.terminate()


async def test_connect_to_server(socketio_server):
    """Test connecting to the server."""

    class ImJoyPlugin:
        """Represent a test plugin."""

        def __init__(self, ws):
            self._ws = ws

        async def setup(self):
            """Set up the plugin."""
            await self._ws.log("initialized")

        async def run(self, ctx):
            """Run the plugin."""
            await self._ws.log("hello world")

    with pytest.raises(Exception, match=r".*Workspace test does not exist.*"):
        ws = await connect_to_server(
            {"name": "my plugin", "workspace": "test", "server_url": SERVER_URL}
        )
    ws = await connect_to_server({"name": "my plugin", "server_url": SERVER_URL})
    await ws.export(ImJoyPlugin(ws))

    ws = await connect_to_server({"server_url": SERVER_URL})
    assert len(ws.config.name) == 36


def test_plugin_runner(socketio_server):
    """Test the plugin runner."""
    with subprocess.Popen(
        [
            sys.executable,
            "-m",
            "imjoy.runner",
            f"--server-url=http://127.0.0.1:{PORT}",
            "--quit-on-ready",
            os.path.join(os.path.dirname(__file__), "example_plugin.py"),
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as proc:
        out, err = proc.communicate()
        assert err.decode("utf8") == ""
        output = out.decode("utf8")
        assert "Generated token: imjoy@" in output
        assert "echo: a message" in output


async def test_plugin_runner_workspace(socketio_server):
    """Test the plugin runner with workspace."""
    api = await connect_to_server(
        {"name": "my second plugin", "server_url": SERVER_URL}
    )
    ret = await api.generate_token()
    assert "id" in ret and "token" in ret
    assert ret["token"].startswith("imjoy@")

    # The following code without passing the token should fail
    # Here we assert the output message contains "permission denied"
    with subprocess.Popen(
        [
            sys.executable,
            "-m",
            "imjoy.runner",
            f"--server-url=http://127.0.0.1:{PORT}",
            f"--workspace={api.config['workspace']}",
            # f"--token={ret['token']}",
            "--quit-on-ready",
            os.path.join(os.path.dirname(__file__), "example_plugin.py"),
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as proc:
        out, err = proc.communicate()
        assert proc.returncode == 1
        assert err.decode("utf8") == ""
        output = out.decode("utf8")
        assert "Permission denied for workspace:" in output

    # now with the token, it should pass
    with subprocess.Popen(
        [
            sys.executable,
            "-m",
            "imjoy.runner",
            f"--server-url=http://127.0.0.1:{PORT}",
            f"--workspace={api.config['workspace']}",
            f"--token={ret['token']}",
            "--quit-on-ready",
            os.path.join(os.path.dirname(__file__), "example_plugin.py"),
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as proc:
        out, err = proc.communicate()
        assert proc.returncode == 0
        assert err.decode("utf8") == ""
        output = out.decode("utf8")
        assert "Generated token: imjoy@" in output
        assert "echo: a message" in output


async def test_workspace(socketio_server):
    """Test the plugin runner."""
    api = await connect_to_server({"name": "my plugin", "server_url": SERVER_URL})
    with pytest.raises(
        Exception, match=r".*Scopes must be empty or contains only the workspace name*"
    ):
        await api.generate_token({"scopes": ["test-workspace"]})
    ret = await api.generate_token()
    assert "id" in ret and "token" in ret
    assert ret["token"].startswith("imjoy@")

    ws = await api.create_workspace(
        {
            "name": "test-workspace",
            "owners": ["user1@imjoy.io", "user2@imjoy.io"],
            "allow_list": [],
            "deny_list": [],
            "visibility": "protected",  # or public
        }
    )
    await ws.log("hello")
    await ws.register_service(
        {
            "name": "test_service",
            "type": "#test",
        }
    )
    service = await ws.get_services({"type": "#test"})
    assert len(service) == 1

    # we should not get it because api is in another workspace
    ss2 = await api.get_services({"type": "#test"})
    assert len(ss2) == 0

    # let's generate a token for the test-workspace
    ret = await ws.generate_token()
    token = ret["token"]

    # now if we connect directly to the workspace
    # we should be able to get the test-workspace services
    api2 = await connect_to_server(
        {
            "name": "my plugin 2",
            "workspace": "test-workspace",
            "server_url": SERVER_URL,
            "token": token,
        }
    )
    assert api2.config["workspace"] == "test-workspace"
    await api2.export({"foo": "bar"})
    ss3 = await api2.get_services({"type": "#test"})
    assert len(ss3) == 1

    plugin = await api2.get_plugin("my plugin 2")
    assert plugin.foo == "bar"

    await api2.export({"foo2": "bar2"})
    plugin = await api2.get_plugin("my plugin 2")
    assert plugin.foo is None
    assert plugin.foo2 == "bar2"

    with pytest.raises(Exception, match=r".*Plugin my plugin 2 not found.*"):
        await api.get_plugin("my plugin 2")

    with pytest.raises(
        Exception, match=r".*Workspace authorizer is not supported yet.*"
    ):
        await api.create_workspace(
            {
                "name": "my-workspace",
                "owners": ["user1@imjoy.io", "user2@imjoy.io"],
                "allow_list": [],
                "deny_list": [],
                "visibility": "protected",  # or public
                "authorizer": "my-plugin::my_authorizer",
            }
        )

    ws2 = await api.get_workspace("test-workspace")
    assert ws.config == ws2.config

    await ws2.set({"docs": "https://imjoy.io"})
    with pytest.raises(Exception, match=r".*Changing workspace name is not allowed.*"):
        await ws2.set({"name": "new-name"})

    with pytest.raises(Exception):
        await ws2.set({"covers": [], "non-exist-key": 999})



"""Set up the imjoy engine package."""
import json
from pathlib import Path

from setuptools import find_packages, setup

DESCRIPTION = (
    "ImJoy Plugin Engine for running Python plugins locally "
    "or remotely from ImJoy.io"
)

try:
    # for Google Colab
    # pylint: disable=unused-import
    import google.colab.output  # noqa: F401

    REQUIREMENTS = ["numpy", "imjoy-rpc>=0.3.0", "imjoy-elfinder"]
except ImportError:
    REQUIREMENTS = [
        "numpy",
        "imjoy-rpc>=0.3.11",
        "pydantic[email]>=1.8.1",
        "typing-extensions>=3.7.4.3",  # required by pydantic
        "python-dotenv>=0.17.0",
        "python-engineio==4.0.0",
        "python-jose==3.2.0",
        "python-socketio[asyncio_client]==5.0.4",
        "pyyaml",
        "fastapi>=0.63.0",
        "uvicorn>=0.13.4",
    ]

ROOT_DIR = Path(__file__).parent.resolve()
README_FILE = ROOT_DIR / "README.md"
LONG_DESCRIPTION = README_FILE.read_text(encoding="utf-8")
VERSION_FILE = ROOT_DIR / "imjoy" / "VERSION"
VERSION = json.loads(VERSION_FILE.read_text())["version"]


setup(
    name="imjoy",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="http://github.com/imjoy-team/ImJoy-Engine",
    author="ImJoy Team",
    author_email="imjoy.team@gmail.com",
    license="MIT",
    packages=find_packages(),
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=REQUIREMENTS,
    extras_require={
        "jupyter": ["jupyter>=1.0.0", "ipykernel>=5.1.4", "imjoy-jupyter-extension"],
    },
    zip_safe=False,
    entry_points={"console_scripts": ["imjoy = imjoy.__main__:main"]},
)



"""Provide a plugin interface."""
import asyncio
import logging
import sys
import uuid

from imjoy_rpc.rpc import RPC
from imjoy_rpc.utils import ContextLocal, dotdict

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("dynamic-plugin")
logger.setLevel(logging.INFO)


class DynamicPlugin:
    """Represent a dynamic plugin."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, config, interface, connection, workspace):
        """Set up instance."""
        self.loop = asyncio.get_event_loop()
        self.config = dotdict(config)
        assert self.config.workspace == workspace.name
        self.workspace = workspace
        self.id = self.config.id or str(uuid.uuid4())  # pylint: disable=invalid-name
        self.name = self.config.name
        self.initializing = False
        self._disconnected = True
        self._log_history = []
        self.connection = connection
        self.authorizer = None
        self.api = None
        self.running = False
        self.terminating = False

        # Note: we don't need to bind the interface
        # to the plugin as we do in the js version
        # We will use context variables `current_plugin`
        # to obtain the current plugin
        self._initial_interface = dotdict(interface)
        self._initial_interface._intf = True
        self._initial_interface.config = self.config.copy()
        if "token" in self._initial_interface.config:
            del self._initial_interface.config["token"]
        self.initialize_if_needed(self.connection, self.config)

        def initialized(data):
            """Handle initialized message."""
            if "error" in data:
                self.error(data["error"])
                logger.error("Plugin failed to initialize: %s", data["error"])
                raise Exception(data["error"])

            asyncio.ensure_future(self._setup_rpc(connection, data["config"]))

        self.connection.on("initialized", initialized)
        self.connection.connect()

    async def _setup_rpc(self, connection, plugin_config):
        """Set up rpc."""
        self.initializing = True
        logger.info("Setting up imjoy-rpc for %s", plugin_config["name"])
        _rpc_context = ContextLocal()
        _rpc_context.api = self._initial_interface
        _rpc_context.default_config = {}
        self._rpc = RPC(connection, _rpc_context)
        self._register_rpc_events()
        self._rpc.set_interface(self._initial_interface)
        await self._send_interface()
        self._allow_execution = plugin_config.get("allow_execution")
        if self._allow_execution:
            await self._execute_plugin()

        self.config.passive = self.config.passive or plugin_config.get("passive")
        if self.config.passive:

            async def func(*args):
                pass

            self.api = dotdict(
                passive=True, _rintf=True, setup=func, on=func, off=func, emit=func
            )
        else:
            self.api = await self._request_remote()

        self.api["config"] = dotdict(
            id=self.id,
            name=self.config.name,
            namespace=self.config.namespace,
            type=self.config.type,
            workspace=self.config.workspace,
            tag=self.config.tag,
        )

        self._disconnected = False
        self.initializing = False
        logger.info(
            "Plugin loaded successfully (workspace=%s, "
            "name=%s, description=%s, api=%s)",
            self.config.workspace,
            self.name,
            self.config.description,
            list(self.api),
        )

    def error(self, *args):
        """Log an error."""
        self._log_history.append({"type": "error", "value": args})
        logger.error("Error in Plugin %s: $%s", self.id, args)

    def log(self, *args):
        """Log."""
        if isinstance(args[0], dict):
            self._log_history.append(args[0])
            logger.info("Plugin $%s:%s", self.id, args[0])
        else:
            msg = " ".join(map(str, args))
            self._log_history.append({"type": "info", "value": msg})
            logger.info("Plugin $%s: $%s", self.id, msg)

    def _set_disconnected(self):
        """Set disconnected state."""
        self._disconnected = True
        self.running = False
        self.initializing = False
        self.terminating = False

    def _register_rpc_events(self):
        """Register rpc events."""

        def disconnected(details):
            """Handle disconnected."""
            if details:
                if "error" in details:
                    self.error(details["message"])
                if "info" in details:
                    self.log(details.info)
            self._set_disconnected()

        self._rpc.on("disconnected", disconnected)

        def remote_ready(_):
            """Handle remote ready."""
            api = self._rpc.get_remote()
            # this make sure if reconnect, setup will be called again
            if "setup" in api:
                asyncio.ensure_future(api.setup())

        self._rpc.on("remoteReady", remote_ready)

        def remote_idle():
            """Handle remote idle."""
            self.running = False

        self._rpc.on("remoteIdle", remote_idle)

        def remote_busy():
            """Handle remote busy."""
            self.running = True

        self._rpc.on("remoteBusy", remote_busy)

    async def _execute_plugin(self):
        """Execute plugin."""
        raise NotImplementedError

    def _send_interface(self):
        """Send the interface."""
        fut = self.loop.create_future()

        def interface_set_as_remote(result):
            """Set interface as remote."""
            fut.set_result(result)

        # pylint: disable=protected-access
        self._rpc._connection.once("interfaceSetAsRemote", interface_set_as_remote)
        self._rpc.send_interface()
        return fut

    def _request_remote(self):
        """Request remote."""
        fut = self.loop.create_future()

        def remote_ready(result):
            """Set remote ready."""
            try:
                fut.set_result(self._rpc.get_remote())
            # TODO: this happens when the the plugin is reconnected
            except asyncio.InvalidStateError:
                pass

        self._rpc.once("remoteReady", remote_ready)
        self._rpc.request_remote()
        return fut

    @staticmethod
    def initialize_if_needed(connection, default_config):
        """Initialize if needed."""

        def imjoy_rpc_ready(data):
            """Handle rpc ready message."""
            config = data["config"] or {}
            forwarding_functions = ["close", "on", "off", "emit"]
            type_ = config.get("type") or default_config.get("type")
            if type_ in ["rpc-window", "window"]:
                forwarding_functions = forwarding_functions + [
                    "resize",
                    "show",
                    "hide",
                    "refresh",
                ]

            credential = None
            if config.get("credential_required"):
                if isinstance(config.credential_fields, list):
                    raise Exception(
                        "Please specify the `config.credential_fields` "
                        "as an array of object."
                    )

                if default_config["credential_handler"]:
                    credential = default_config["credential_handler"](
                        config["credential_fields"]
                    )

                else:
                    credential = {}
                    # for k in config['credential_fields']:
                    #     credential[k.id] = prompt(k.label, k.value)
            connection.emit(
                {
                    "type": "initialize",
                    "config": {
                        "name": default_config.get("name"),
                        "type": default_config.get("type"),
                        "allow_execution": True,
                        "enable_service_worker": True,
                        "forwarding_functions": forwarding_functions,
                        "expose_api_globally": True,
                        "credential": credential,
                    },
                    "peer_id": data["peer_id"],
                }
            )

        connection.once("imjoyRPCReady", imjoy_rpc_ready)

    async def terminate(self, force=False):
        """Terminate."""
        try:
            if self.api and self.api.exit and callable(self.api.exit):
                logger.info(
                    "Terminating plugin %s/%s", self.config.workspace, self.name
                )
                self.api.exit()
        finally:
            logger.info("Plugin %s terminated.", self.config.name)
            self._set_disconnected()



"""Provide authentication."""
import json
import logging
import ssl
import sys
import time
import traceback
import uuid
from os import environ as env
from typing import List, Optional
from urllib.request import urlopen

from dotenv import find_dotenv, load_dotenv
from fastapi import Header, HTTPException, Request
from jose import jwt
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from imjoy.core import UserInfo, VisibilityEnum, TokenConfig, all_users, all_workspaces

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("imjoy-core")
logger.setLevel(logging.INFO)

ENV_FILE = find_dotenv()
if ENV_FILE:
    load_dotenv(ENV_FILE)

AUTH0_DOMAIN = env.get("AUTH0_DOMAIN", "imjoy.eu.auth0.com")
AUTH0_AUDIENCE = env.get("AUTH0_AUDIENCE", "https://imjoy.eu.auth0.com/api/v2/")
JWT_SECRET = env.get("JWT_SECRET")
if not JWT_SECRET:
    logger.warning("JWT_SECRET is not defined")
    JWT_SECRET = str(uuid.uuid4())


class AuthError(Exception):
    """Represent an authentication error."""

    def __init__(self, error, status_code):
        """Set up instance."""
        super().__init__()
        self.error = error
        self.status_code = status_code


class ValidToken(BaseModel):
    """Represent a valid token."""

    credentials: dict
    scopes: List[str] = []

    def has_scope(self, checked_token):
        """Return True if the token has the correct scope."""
        if checked_token in self.scopes:
            return True

        raise HTTPException(
            status_code=403, detail="Not authorized to perform this action"
        )


def login_required(request: Request, authorization: str = Header(None)):
    """Return token if login is ok."""
    return valid_token(authorization, request)


def admin_required(request: Request, authorization: str = Header(None)):
    """Return token if the token has an admin role."""
    token = valid_token(authorization, request)
    roles = token.credentials.get("https://api.imjoy.io/roles", [])
    if "admin" not in roles:
        raise HTTPException(status_code=401, detail="Admin required")
    return token


def is_admin(token):
    """Check if token has an admin role."""
    roles = token.credentials.get("https://api.imjoy.io/roles", [])
    if "admin" not in roles:
        return False
    return True


def get_user_email(token):
    """Return the user email from the token."""
    return token.credentials.get("https://api.imjoy.io/email")


def get_user_id(token):
    """Return the user id from the token."""
    return token.credentials.get("sub")


def get_user_info(token):
    """Return the user info from the token."""
    return {
        "user_id": token.credentials.get("sub"),
        "email": token.credentials.get("https://api.imjoy.io/email"),
        "roles": token.credentials.get("https://api.imjoy.io/roles", []),
    }


JWKS = None


def get_rsa_key(kid, refresh=False):
    """Return an rsa key."""
    global JWKS  # pylint: disable=global-statement
    if JWKS is None or refresh:
        with urlopen(
            f"https://{AUTH0_DOMAIN}/.well-known/jwks.json",
            # pylint: disable=protected-access
            context=ssl._create_unverified_context(),
        ) as jsonurl:
            JWKS = json.loads(jsonurl.read())
    rsa_key = {}
    for key in JWKS["keys"]:
        if key["kid"] == kid:
            rsa_key = {
                "kty": key["kty"],
                "kid": key["kid"],
                "use": key["use"],
                "n": key["n"],
                "e": key["e"],
            }
            break
    return rsa_key


def simulate_user_token(returned_token, request):
    """Allow admin all_users to simulate another user."""
    if "user_id" in request.query_params:
        returned_token.credentials["sub"] = request.query_params["user_id"]
    if "email" in request.query_params:
        returned_token.credentials["https://api.imjoy.io/email"] = request.query_params[
            "email"
        ]
    if "roles" in request.query_params:
        returned_token.credentials["https://api.imjoy.io/roles"] = request.query_params[
            "roles"
        ].split(",")


def valid_token(authorization: str, request: Optional[Request] = None):
    """Validate token."""
    if not authorization:
        raise HTTPException(status_code=401, detail="Authorization header is expected")

    parts = authorization.split()

    if parts[0].lower() != "bearer":
        raise HTTPException(
            status_code=401, detail="Authorization header must start with" " Bearer"
        )
    if len(parts) == 1:
        raise HTTPException(status_code=401, detail="Token not found")
    if len(parts) > 2:
        raise HTTPException(
            status_code=401, detail="Authorization header must be 'Bearer' token"
        )

    authorization = parts[1]
    try:
        unverified_header = jwt.get_unverified_header(authorization)

        # Get RSA key
        rsa_key = get_rsa_key(unverified_header["kid"], refresh=False)
        # Try to refresh jwks if failed
        if not rsa_key:
            rsa_key = get_rsa_key(unverified_header["kid"], refresh=True)

        # Decode token
        payload = jwt.decode(
            authorization,
            rsa_key,
            algorithms=["RS256"],
            audience=AUTH0_AUDIENCE,
            issuer=f"https://{AUTH0_DOMAIN}/",
        )

        returned_token = ValidToken(
            credentials=payload, scopes=payload["scope"].split(" ")
        )

        # This is needed for patching the test token
        if "create:roles" in payload["scope"]:
            if "https://api.imjoy.io/roles" not in returned_token.credentials:
                returned_token.credentials["https://api.imjoy.io/roles"] = ["admin"]
            if "https://api.imjoy.io/email" not in returned_token.credentials:
                returned_token.credentials["https://api.imjoy.io/email"] = None

        if (
            "admin" in returned_token.credentials["https://api.imjoy.io/roles"]
            and request
        ):
            simulate_user_token(returned_token, request)

        return returned_token

    except jwt.ExpiredSignatureError as err:
        raise HTTPException(
            status_code=401, detail="The token has expired. Please fetch a new one"
        ) from err
    except jwt.JWTError as err:
        raise HTTPException(status_code=401, detail=traceback.format_exc()) from err


def parse_token(authorization):
    """Parse the token."""
    parts = authorization.split()
    if parts[0].lower() != "bearer":
        raise Exception("Authorization header must start with" " Bearer")
    if len(parts) == 1:
        raise Exception("Token not found")
    if len(parts) > 2:
        raise Exception("Authorization header must be 'Bearer' token")

    token = parts[1]
    if not token.startswith("imjoy@"):
        # auth0 token
        return get_user_info(valid_token(authorization))
    # generated token
    token = token.lstrip("imjoy@")
    return jwt.decode(token, JWT_SECRET, algorithms=["HS256"])


def generate_presigned_token(user_info: UserInfo, config: TokenConfig):
    """Generate presigned tokens.

    This will generate a token which will be connected as a child user.
    Child user may generate more child user token if it has admin permission.
    """
    scopes = config.scopes
    for scope in scopes:
        if not check_permission(scope, user_info):
            raise PermissionError(f"User has no permission to scope: {scope}")

    # always generate a new user id
    uid = str(uuid.uuid4())
    expires_in = config.expires_in
    if expires_in:
        expires_at = time.time() + expires_in
    else:
        expires_at = None
    token = jwt.encode(
        {
            "scopes": scopes,
            "expires_at": expires_at,
            "user_id": uid,
            "parent": user_info.parent if user_info.parent else user_info.id,
            "email": config.email,
            "roles": [],
        },
        JWT_SECRET,
        algorithm="HS256",
    )
    return {"token": "imjoy@" + token, "id": uid}


def check_permission(workspace, user_info):
    """Check user permission for a workspace."""
    # pylint: disable=too-many-return-statements
    if isinstance(workspace, str):
        workspace = all_workspaces.get(workspace)
        if not workspace:
            logger.warning("Workspace %s not found", workspace)
            return False

    if workspace.name == user_info.id:
        return True

    if user_info.parent:
        parent = all_users.get(user_info.parent)
        if not parent:
            return False
        if not check_permission(workspace, parent):
            return False
        # if the parent has access
        # and the workspace is in the scopes
        # then we allow the access
        if workspace.name in user_info.scopes:
            return True

    _id = user_info.email or user_info.id

    if _id in workspace.owners:
        return True

    if workspace.visibility == VisibilityEnum.public:
        if workspace.deny_list and user_info.email not in workspace.deny_list:
            return True
    elif workspace.visibility == VisibilityEnum.protected:
        if workspace.allow_list and user_info.email in workspace.allow_list:
            return True

    if "admin" in user_info.roles:
        logger.info(
            "Allowing access to %s for admin user %s", workspace.name, user_info.id
        )
        return True

    return False



"""Provide interface functions for the core."""
import logging
import sys
from functools import partial
from typing import Optional

from imjoy.core import (
    TokenConfig,
    WorkspaceInfo,
    all_workspaces,
    current_plugin,
    current_user,
    current_workspace,
)
from imjoy.core.auth import check_permission, generate_presigned_token

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("imjoy-core")
logger.setLevel(logging.INFO)


class CoreInterface:
    """Represent the interface of the ImJoy core."""

    # pylint: disable=no-self-use, protected-access

    def __init__(self, imjoy_api=None):
        """Set up instance."""
        self.imjoy_api = imjoy_api

    def register_service(self, service: dict):
        """Register a service."""
        plugin = current_plugin.get()
        workspace = current_workspace.get()
        service.provider = plugin.name
        service.providerId = plugin.id
        service._rintf = True
        workspace._services.append(service)

    def get_plugin(self, name):
        """Return a plugin by its name."""
        workspace = current_workspace.get()

        if name in workspace._plugins:
            return workspace._plugins[name].api
        raise Exception(f"Plugin {name} not found")

    def get_services(self, query: dict):
        """Return a list of services based on the query."""
        workspace = current_workspace.get()
        ret = []
        for service in workspace._services:
            match = True
            for key in query:
                if service[key] != query[key]:
                    match = False
            if match:
                ret.append(service)
        return ret

    def log(self, msg):
        """Log a plugin message."""
        plugin = current_plugin.get()
        logger.info("%s: %s", plugin.name, msg)

    def error(self, msg):
        """Log a plugin error message."""
        plugin = current_plugin.get()
        logger.error("%s: %s", plugin.name, msg)

    def generate_token(self, config: Optional[dict] = None):
        """Generate a token for the current workspace."""
        workspace = current_workspace.get()
        config = config or {}
        if "scopes" in config and config["scopes"] != [workspace.name]:
            raise Exception("Scopes must be empty or contains only the workspace name.")
        config["scopes"] = [workspace.name]
        token_config = TokenConfig.parse_obj(config)
        return generate_presigned_token(current_user.get(), token_config)

    def create_workspace(self, config: dict):
        """Create a new workspace."""
        config["persistent"] = config.get("persistent") or False
        workspace = WorkspaceInfo.parse_obj(config)
        if workspace.name in all_workspaces:
            raise Exception(f"Workspace {workspace.name} already exists.")
        if workspace.authorizer:
            raise Exception("Workspace authorizer is not supported yet.")
        user_info = current_user.get()
        # make sure we add the user's email to owners
        _id = user_info.email or user_info.id
        if _id not in workspace.owners:
            workspace.owners.append(_id)
        workspace.owners = [o.strip() for o in workspace.owners if o.strip()]
        user_info.scopes.append(workspace.name)
        all_workspaces[workspace.name] = workspace
        return self.get_workspace(workspace.name)

    def _update_workspace(self, name, config: dict):
        """Bind the context to the generated workspace."""
        if not name:
            raise Exception("Workspace name is not specified.")
        if name not in all_workspaces:
            raise Exception(f"Workspace {name} not found")
        workspace = all_workspaces[name]
        user_info = current_user.get()
        if not check_permission(workspace, user_info):
            raise PermissionError(f"Permission denied for workspace {name}")

        if "name" in config:
            raise Exception("Changing workspace name is not allowed.")

        # make sure all the keys are valid
        # TODO: verify the type
        for key in config:
            if key.startswith("_") or not hasattr(workspace, key):
                raise KeyError(f"Invalid key: {key}")

        for key in config:
            if not key.startswith("_") and hasattr(workspace, key):
                setattr(workspace, key, config[key])
        # make sure we add the user's email to owners
        _id = user_info.email or user_info.id
        if _id not in workspace.owners:
            workspace.owners.append(_id)
        workspace.owners = [o.strip() for o in workspace.owners if o.strip()]

    def get_workspace(self, name: str):
        """Bind the context to the generated workspace."""
        if name not in all_workspaces:
            raise Exception(f"Workspace {name} not found")
        workspace = all_workspaces[name]
        user_info = current_user.get()
        if not check_permission(workspace, user_info):
            raise PermissionError(f"Permission denied for workspace {name}")

        interface = self.get_interface()
        bound_interface = {}
        for key in interface:
            if callable(interface[key]):

                def wrap_func(func, *args, **kwargs):
                    workspace_bk = current_workspace.get()
                    ret = None
                    try:
                        current_workspace.set(workspace)
                        ret = func(*args, **kwargs)
                    except Exception as exp:
                        raise exp
                    finally:
                        current_workspace.set(workspace_bk)
                    return ret

                bound_interface[key] = partial(wrap_func, interface[key])
                bound_interface[key].__name__ = key  # required for imjoy-rpc
            else:
                bound_interface[key] = interface[key]
        bound_interface["config"] = {"workspace": name}
        bound_interface["set"] = partial(self._update_workspace, name)
        bound_interface["_rintf"] = True
        return bound_interface

    def get_interface(self):
        """Return the interface."""
        return {
            "_rintf": True,
            "log": self.log,
            "error": self.error,
            "registerService": self.register_service,
            "register_service": self.register_service,
            "getServices": self.get_services,
            "get_services": self.get_services,
            "utils": {},
            "getPlugin": self.get_plugin,
            "get_plugin": self.get_plugin,
            "generateToken": self.generate_token,
            "generate_token": self.generate_token,
            "create_workspace": self.create_workspace,
            "createWorkspace": self.create_workspace,
            "get_workspace": self.get_workspace,
            "getWorkspace": self.get_workspace,
        }



"""Provide utilities that should not be aware of ImJoy engine."""
import copy
import json
import os
import string
import sys
import threading
import time
import uuid
from importlib import import_module

if sys.platform == "win32":
    from ctypes import windll

    def get_drives():
        """Return windows drives."""
        drives = []
        bitmask = windll.kernel32.GetLogicalDrives()
        for letter in string.ascii_uppercase:
            if bitmask & 1:
                drives.append(os.path.abspath(letter + ":/"))
            bitmask >>= 1
        return drives


_SERVER_THREAD = None


def _show_elfinder_colab(root_dir="/content", port=8765, height=600, width="100%"):
    # pylint: disable=import-error, import-outside-toplevel, no-name-in-module
    from google.colab import output
    from imjoy_elfinder.app import main

    global _SERVER_THREAD  # pylint: disable=global-statement
    if _SERVER_THREAD is None:

        def start_elfinder():
            global _SERVER_THREAD  # pylint: disable=global-statement
            try:
                main(["--root-dir={}".format(root_dir), "--port={}".format(port)])
            except OSError:
                print("ImJoy-elFinder server already started.")
            _SERVER_THREAD = thread

        # start imjoy-elfinder server
        thread = threading.Thread(target=start_elfinder)
        thread.start()

    time.sleep(1)
    output.serve_kernel_port_as_iframe(port, height=str(height), width=str(width))


def _show_elfinder_jupyter(url="/elfinder", height=600, width="100%"):
    from IPython import display  # pylint: disable=import-outside-toplevel

    code = """(async (url, width, height, element) => {
        element.appendChild(document.createTextNode(''));
        const iframe = document.createElement('iframe');
        iframe.src = url;
        iframe.height = height;
        iframe.width = width;
        iframe.style.border = 0;
        element.appendChild(iframe);
        })""" + "({url}, {width}, {height}, element[0])".format(
        url=json.dumps(url), width=json.dumps(width), height=json.dumps(height)
    )
    display.display(display.Javascript(code))


def show_elfinder(**kwargs):
    """Show elfinder."""
    try:
        # pylint: disable=import-outside-toplevel, unused-import
        from google.colab import output  # noqa: F401

        is_colab = True
    except ImportError:
        is_colab = False

    if is_colab:
        _show_elfinder_colab(**kwargs)
    else:
        _show_elfinder_jupyter(**kwargs)


def read_or_generate_token(token_path=None):
    """Read or generate token."""
    token_path = token_path or os.path.join(os.path.expanduser("~"), ".jupyter_token")
    # read token from file if exists
    try:
        with open(token_path, "r") as fil:
            token = fil.read()
    except FileNotFoundError:
        token = str(uuid.uuid4())
        with open(token_path, "w") as fil:
            fil.write(token)

    return token


def write_token(token, token_path=None):
    """Write token."""
    token_path = token_path or os.path.join(os.path.expanduser("~"), ".jupyter_token")
    with open(token_path, "w") as fil:
        fil.write(token)


def parse_repos(requirements, work_dir):
    """Return a list of repositories from a list of requirements."""
    repos = []
    if isinstance(requirements, list):
        requirements = [str(req) for req in requirements]
        for req in requirements:
            if ":" in req:
                req_parts = req.split(":")
                typ, libs = req_parts[0], ":".join(req_parts[1:])
                typ, libs = typ.strip(), libs.strip()
                libs = [lib.strip() for lib in libs.split(" ") if lib.strip() != ""]
                if typ == "repo" and libs:
                    name = libs[0].split("/")[-1].replace(".git", "")
                    repo = {
                        "url": libs[0],
                        "repo_dir": os.path.join(
                            work_dir, libs[1] if len(libs) > 1 else name
                        ),
                    }
                    repos.append(repo)
    return repos


def console_to_str(string_):
    """From pypa/pip project, pip.backwardwardcompat. License MIT."""
    try:
        return string_.decode(sys.__stdout__.encoding)
    except UnicodeDecodeError:
        return string_.decode("utf_8")
    except AttributeError:  # for tests, #13
        return string_


class dotdict(dict):  # pylint: disable=invalid-name
    """Access dictionary attributes with dot.notation."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __deepcopy__(self, memo=None):
        """Make a deep copy."""
        return dotdict(copy.deepcopy(dict(self), memo=memo))


def get_psutil():
    """Try to import and return psutil."""
    try:
        return import_module("psutil")
    except ImportError:
        print(
            "WARNING: a library called 'psutil' can not be imported, "
            "this may cause problem when killing processes."
        )
        return None


def kill_process(pid, logger=None):
    """Kill process."""
    psutil = get_psutil()
    if psutil is None:
        return
    if logger:
        logger.info("Killing process (pid=%s)", pid)
    try:
        current_process = psutil.Process(pid)
        for proc in current_process.children(recursive=True):
            try:
                if proc.is_running():
                    proc.kill()
            except psutil.NoSuchProcess:
                if logger:
                    logger.info("Subprocess %s has already been killed", pid)
            except Exception as exc:  # pylint: disable=broad-except
                if logger:
                    logger.error(
                        "Failed to kill a subprocess (pid=%s). Error: %s", pid, exc
                    )
        current_process.kill()
        if logger:
            logger.info("Process %s was killed.", pid)
    except psutil.NoSuchProcess:
        if logger:
            logger.info("Process %s has already been killed", pid)
    except Exception as exc:  # pylint: disable=broad-except
        if logger:
            logger.error(
                "Failed to kill a process (pid=%s), "
                "you may want to kill it manually. Error: %s",
                pid,
                exc,
            )


def scandir(path, type_=None, recursive=False):
    """Scan a directory for a type of files return a list of files found."""
    file_list = []
    for fil in os.scandir(path):
        if fil.name.startswith("."):
            continue
        if type_ is None or type_ == "file":
            if os.path.isdir(fil.path):
                if recursive:
                    file_list.append(
                        {
                            "name": fil.name,
                            "type": "dir",
                            "children": scandir(fil.path, type_, recursive),
                        }
                    )
                else:
                    file_list.append({"name": fil.name, "type": "dir"})
            else:
                file_list.append({"name": fil.name, "type": "file"})
        elif type_ == "directory":
            if os.path.isdir(fil.path):
                file_list.append({"name": fil.name})
    return file_list
